use anyhow::{anyhow, Result};

use std::io;

pub use serialport::{DataBits, FlowControl, Parity, SerialPort, StopBits};
pub use std::time::Duration;

pub enum SerialReadResult {
    Message(String),
    Timeout,
}

pub struct SerialPortSettings {
    pub baud_rate: u32,
    pub data_bits: DataBits,
    pub parity: Parity,
    pub stop_bits: StopBits,
    pub flow_control: FlowControl,
    pub timeout: Duration,
}

pub struct Port {
    serial_port: Box<dyn SerialPort>,
}

impl Port {
    pub fn open(settings: SerialPortSettings) -> Result<Self> {
        let port_path = Self::find_macroboard()?;

        let port = serialport::new(port_path, settings.baud_rate)
            .data_bits(settings.data_bits)
            .parity(settings.parity)
            .stop_bits(settings.stop_bits)
            .flow_control(settings.flow_control)
            .timeout(settings.timeout)
            .open()?;

        Ok(Port { serial_port: port })
    }

    #[allow(unused)]
    pub fn read(&mut self) -> Result<SerialReadResult> {
        let mut buf: Vec<u8> = (0..255).collect();

        match self.serial_port.read(&mut buf[..]) {
            Ok(t) => Ok(SerialReadResult::Message(
                String::from_utf8(buf[0..t].to_vec())
                    .map_err(|err| anyhow!("failed converting utf8 buffer ({})", err))?,
            )),
            Err(ref e) if e.kind() == io::ErrorKind::TimedOut => Ok(SerialReadResult::Timeout),
            Err(err) => Err(anyhow!("failed reading serial port ({})", err)),
        }
    }

    #[allow(unused)]
    pub fn write(&mut self, msg: &str) -> Result<()> {
        self.serial_port
            .write(msg.as_bytes())
            .map_err(|err| anyhow!("failed writing to serial port ({})", err))?;

        Ok(())
    }

    fn find_macroboard() -> Result<String> {
        let available_ports = serialport::available_ports()
            .map_err(|err| anyhow!("error querying serial ports ( {})", err))?;

        for available_port in available_ports.iter() {
            if let serialport::SerialPortType::UsbPort(usb_info) = &available_port.port_type {
                // check for the correct device
                if let Some(product_name) = &usb_info.product {
                    if product_name == "FT232R_USB_UART"
                        && usb_info.vid == 0x0403
                        && usb_info.pid == 0x6001
                    {
                        return Ok(available_port.port_name.clone());
                    }
                }
            }
        }

        return Err(anyhow!("macro board not found on usb bus".to_string()));
    }
}
